from django.core.management.base import CommandError
from django.conf import settings
import os
import signal
import sys
import time
import subprocess

class RunOnce(object):
    def __init__(self, attrname):
        self.attrname = attrname
    def __call__(self, f):
        attrname = self.attrname
        def w(self, *args, **kwargs):
            if not hasattr(self, attrname):
                setattr(self, attrname, f(self, *args, **kwargs))
            return getattr(self, attrname)                
        return w
        
class Backend(object):
    def __init__(self, name, graceful):
        self.name = name
        self.graceful = graceful
        self.socket = os.path.join(self.graceful.statedir, "fastcgi-%s.socket" % self.name)
        self.pidfile = os.path.join(self.graceful.statedir, "fastcgi-%s.pid" % self.name)
    
    def is_up(self):
        if os.path.exists(self.pidfile):
            try:
                f = open(self.pidfile, 'r')
                pid = int(f.read())
                f.close()
            except (IOError, ValueError):
                return False
            try:
                os.kill(pid, 0)
                return pid
            except OSError:
                return False
        else:
            return False
            
    def is_active(self):
        try:
            return os.readlink(self.graceful.socket) == self.socket
        except OSError:
            return False
            
    def __repr__(self):
        return "[%s] backend %s (%s)" % (
                '*' if self.is_active() else ' ',
                self.name,
                'running' if self.is_up() else 'down',
            )
            
    def start(self):
        if self.is_up():
            print 'Skipping already running backend "%s"' % self.name
            return
        options = getattr(settings, 'GRACEFUL_OPTIONS', {})
        if type(options) == dict:
            options.update({'socket': self.socket, 'pidfile': self.pidfile, 'daemonize': 'true'})
        else:
            raise CommandError('GRACEFUL_OPTIONS optional setting should be set to a dictionary of options to runfcgi command.')
        subprocess.Popen([sys.executable, sys.argv[0], 'runfcgi'] + ['%s=%s' % (k, options[k]) for k in options]).wait()
        if not self.is_up():
            raise CommandError('Backend "%s" failed to start.' % self.name)
        while not os.path.exists(self.socket):
            time.sleep(0.1)
        print 'Started backend "%s".' % self.name
        try:
            os.chmod(self.socket, 0666)
        except Exception, e:
            print "chmod 666 %s failed" % self.socket
        
    def stop(self):
        pid = self.is_up()
        if not pid:
            print 'Skipping stopped backend "%s"' % self.name
            return
        try:
            os.kill(pid, signal.SIGKILL)
        except Exception, e:
            print "unable to kill %d: %s" % (pid, e)
        for filename in (self.socket, self.pidfile):
            try:
                os.unlink(filename)
            except:
                pass
        print 'Stopped backend "%s".' % self.name
        
    def switch(self):
        if os.path.exists(self.graceful.socket) or os.path.islink(self.graceful.socket):
            os.unlink(self.graceful.socket)
        os.symlink(self.socket, self.graceful.socket)
        print 'Switched to backend "%s".' % self.name
            
class Graceful(object):
    def __init__(self):
        try:
            self.statedir = settings.GRACEFUL_STATEDIR
        except AttributeError:
            raise CommandError('GRACEFUL_STATEDIR setting is required for this command to work.')
        if not os.path.isdir(self.statedir):
            raise CommandError('Path "%s" specified in GRACEFUL_STATEDIR setting should be a writable directory.' % self.statedir)            
        self.socket = os.path.join(self.statedir, "fastcgi.socket")
        self.backends = (
                Backend('a', graceful=self),
                Backend('b', graceful=self),
            )
        self.map = dict([(b.name, b) for b in self.backends])



from django_graceful import Graceful
from django.core.management.base import BaseCommand, CommandError

class Command(BaseCommand):
    help = "Switches to specified backend."

    def handle(self, *args, **options):
        graceful = Graceful()
        if len(args) > 1 or not args:
            raise CommandError('You should specify ONE backend name to switch to.')
        try:
            graceful.map[args[0]].switch()
        except KeyError:
            raise CommandError('Backend "%s" does not exist.' % a)




from django_graceful import Graceful
from django.core.management.base import BaseCommand, CommandError
from optparse import make_option

class Command(BaseCommand):
    help = "Stops specified backends."

    option_list = (
        make_option('--all', help='Stop all running backends', action='store_true'),
        make_option('--inactive', help='Stop all inactive backends', action='store_true'),
    ) + BaseCommand.option_list
    
    def handle(self, *args, **options):
        graceful = Graceful()
        all = options['all']
        inactive = options['inactive']
        if not (args and (not all) and (not inactive) or (not args) and all and (not inactive) or (not args) and (not all) and inactive):
            raise CommandError('Specify backend names to stop or give --all to stop every running backend or --inactive to stop all inactive running backends.')
        stoplist = []
        if all:
            stoplist = [b for b in graceful.backends if b.is_up()]
        elif inactive:
            stoplist = [b for b in graceful.backends if b.is_up() and (not b.is_active())]
        else:
            for a in args:
                try:
                    stoplist.append(graceful.map[a])
                except KeyError:
                    raise CommandError('Backend "%s" does not exist.' % a)
        for b in stoplist:
            b.stop()




from django_graceful import Graceful
from django.core.management.base import NoArgsCommand

class Command(NoArgsCommand):
    help = "Restarts inactive backend and switches to it."
    
    def handle_noargs(self, **options):
        graceful = Graceful()
        j = 0
        for i in range(len(graceful.backends)):
            if graceful.backends[i].is_active():
                j = (i + 1) % len(graceful.backends)
                break
        if graceful.backends[j].is_up():
            graceful.backends[j].stop()
        graceful.backends[j].start()
        if not graceful.backends[j].is_active():
            graceful.backends[j].switch()




from django_graceful import Graceful
from django.core.management.base import NoArgsCommand

class Command(NoArgsCommand):
    help = "Ensures there is a running backend and it's active."
    
    def handle_noargs(self, **options):
        graceful = Graceful()
        up = []
        for b in graceful.backends:
            if b.is_up():
                if b.is_active():
                    return
                else:
                    up.append(b)
        if up:
            up[0].switch()
        else:
            graceful.backends[0].start()
            if not graceful.backends[0].is_active():
                graceful.backends[0].switch()




from django_graceful import Graceful
from django.core.management.base import BaseCommand, CommandError
from optparse import make_option

class Command(BaseCommand):
    help = "Starts specified backends."

    option_list = (
        make_option('--all', help='Start all available backends', action='store_true'),
    ) + BaseCommand.option_list
    
    def handle(self, *args, **options):
        graceful = Graceful()
        all = options['all']
        if not (args and (not all) or (not args) and all):
            raise CommandError('Specify backend names to start or give --all to start everyone available.')
        startlist = []
        if all:
            startlist = list(graceful.backends)
        else:
            for a in args:
                try:
                    startlist.append(graceful.map[a])
                except KeyError:
                    raise CommandError('Backend "%s" does not exist.' % a)
        for b in startlist:
            b.start()




from django_graceful import Graceful
from django.core.management.base import NoArgsCommand

class Command(NoArgsCommand):
    help = "Shows backends list with running/active status."
    
    def handle_noargs(self, **options):
        for b in Graceful().backends:
            print b




#!/usr/bin/env python

from distutils.core import setup

setup(
        name='django_graceful',
        version='0.1',
        description='Django fastcgi deployment tool for easy management and graceful code updating',
        author='Andrey Bulgakov',
        author_email='mail@andreiko.ru',
        url='http://github.com/andreiko/django_graceful',
        packages=['django_graceful', 'django_graceful.management', 'django_graceful.management.commands'],
    )
